from sqlglot.dialects.presto import Presto as presto
from sqlglot import exp
from sqlglot.helper import seq_get
from sqlglot.errors import ParseError
from sqlglot.dialects.dialect import locate_to_strposition
from sqlglot.tokens import TokenType


def _build_approx_percentile(args: list) -> exp.Expression:
    if len(args) == 4:
        arg3 = seq_get(args, 3)
        try:
            number = float(arg3.this) if arg3 is not None else 0
            return exp.ApproxQuantile(
                this=seq_get(args, 0),
                weight=seq_get(args, 1),
                quantile=seq_get(args, 2),
                accuracy=exp.Literal(this=f'{int((1/number) * 100)} ', is_string=False),
            )
        except ValueError as exc:
            raise ParseError(f"Expected a string representation of a number for argument 2, but got {arg3}") from exc
    if len(args) == 3:
        arg2 = seq_get(args, 2)
        try:
            number = float(arg2.this) if arg2 is not None else 0
            return exp.ApproxQuantile(
                this=seq_get(args, 0),
                quantile=seq_get(args, 1),
                accuracy=exp.Literal(this=f'{int((1/number) * 100)}', is_string=False),
            )
        except ValueError as exc:
            raise ParseError(f"Expected a string representation of a number for argument 2, but got {arg2}") from exc
    return exp.ApproxQuantile.from_arg_list(args)


class Presto(presto):

    class Parser(presto.Parser):
        VALUES_FOLLOWED_BY_PAREN = False

        FUNCTIONS = {
            **presto.Parser.FUNCTIONS,
            "APPROX_PERCENTILE": _build_approx_percentile,
            "STRPOS": locate_to_strposition,
        }

    class Tokenizer(presto.Tokenizer):
        KEYWORDS = {
            **presto.Tokenizer.KEYWORDS,
            "JSON": TokenType.TEXT,
        }



import logging
from pathlib import Path

from sqlglot.dialects.dialect import Dialect

from databricks.labs.remorph.snow.sql_transpiler import SqlglotEngine

logger = logging.getLogger(__name__)


class EngineAdapter:
    def __init__(self, dialect: Dialect):
        self.dialect = dialect

    def select_engine(self, input_type: str):
        if input_type.lower() not in {"sqlglot"}:
            msg = f"Unsupported input type: {input_type}"
            logger.error(msg)
            raise ValueError(msg)
        return SqlglotEngine(self.dialect)

    def parse_sql_content(self, dag, sql_content: str, file_name: str | Path, engine: str):
        # Not added type hints for dag as it is a cyclic import
        parser = self.select_engine(engine)
        for root_table, child in parser.parse_sql_content(sql_content, file_name):
            dag.add_node(child)
            dag.add_edge(root_table, child)



import logging
import os
from pathlib import Path

from sqlglot.dialects.dialect import Dialect
from databricks.labs.remorph.__about__ import __version__
from databricks.labs.remorph.config import (
    MorphConfig,
    get_dialect,
    TranspilationResult,
    ValidationResult,
)
from databricks.labs.remorph.helpers import db_sql
from databricks.labs.remorph.helpers.execution_time import timeit
from databricks.labs.remorph.helpers.file_utils import (
    dir_walk,
    is_sql_file,
    make_dir,
    remove_bom,
)
from databricks.labs.remorph.helpers.morph_status import (
    MorphStatus,
    ParserError,
    ValidationError,
)
from databricks.labs.remorph.helpers.validation import Validator
from databricks.labs.remorph.snow import lca_utils
from databricks.labs.remorph.snow.sql_transpiler import SqlglotEngine
from databricks.sdk import WorkspaceClient

# pylint: disable=unspecified-encoding

logger = logging.getLogger(__name__)


def _process_file(
    config: MorphConfig,
    validator: Validator | None,
    transpiler: SqlglotEngine,
    input_file: str | Path,
    output_file: str | Path,
):
    logger.info(f"started processing for the file ${input_file}")
    validate_error_list = []
    no_of_sqls = 0

    input_file = Path(input_file)
    output_file = Path(output_file)

    with input_file.open("r") as f:
        sql = remove_bom(f.read())

    lca_error = lca_utils.check_for_unsupported_lca(get_dialect(config.source.lower()), sql, str(input_file))

    if lca_error:
        validate_error_list.append(lca_error)

    write_dialect = config.get_write_dialect()

    transpiler_result: TranspilationResult = _parse(transpiler, write_dialect, sql, input_file, [])

    with output_file.open("w") as w:
        for output in transpiler_result.transpiled_sql:
            if output:
                no_of_sqls = no_of_sqls + 1
                if config.skip_validation:
                    w.write(output)
                    w.write("\n;\n")
                elif validator:
                    validation_result: ValidationResult = _validation(validator, config, output)
                    w.write(validation_result.validated_sql)
                    if validation_result.exception_msg is not None:
                        validate_error_list.append(ValidationError(str(input_file), validation_result.exception_msg))
            else:
                warning_message = (
                    f"Skipped a query from file {input_file!s}. "
                    f"Check for unsupported operations related to STREAM, TASK, SESSION etc."
                )
                logger.warning(warning_message)

    return no_of_sqls, transpiler_result.parse_error_list, validate_error_list


def _process_directory(
    config: MorphConfig,
    validator: Validator | None,
    transpiler: SqlglotEngine,
    root: str | Path,
    base_root: str,
    files: list[str],
):
    output_folder = config.output_folder
    parse_error_list = []
    validate_error_list = []
    counter = 0

    root = Path(root)

    for file in files:
        logger.info(f"Processing file :{file}")
        if is_sql_file(file):
            if output_folder in {None, "None"}:
                output_folder_base = f"{root.name}/transpiled"
            else:
                output_folder_base = f'{str(output_folder).rstrip("/")}/{base_root}'

            output_file_name = Path(output_folder_base) / Path(file).name
            make_dir(output_folder_base)

            no_of_sqls, parse_error, validation_error = _process_file(
                config, validator, transpiler, file, output_file_name
            )
            counter = counter + no_of_sqls
            parse_error_list.extend(parse_error)
            validate_error_list.extend(validation_error)
        else:
            # Only SQL files are processed with extension .sql or .ddl
            pass

    return counter, parse_error_list, validate_error_list


def _process_recursive_dirs(
    config: MorphConfig, input_sql_path: Path, validator: Validator | None, transpiler: SqlglotEngine
):
    input_sql = input_sql_path
    parse_error_list = []
    validate_error_list = []

    file_list = []
    counter = 0
    for root, _, files in dir_walk(input_sql):
        base_root = str(root).replace(str(input_sql), "")
        folder = str(input_sql.resolve().joinpath(base_root))
        msg = f"Processing for sqls under this folder: {folder}"
        logger.info(msg)
        file_list.extend(files)
        no_of_sqls, parse_error, validation_error = _process_directory(
            config, validator, transpiler, root, base_root, files
        )
        counter = counter + no_of_sqls
        parse_error_list.extend(parse_error)
        validate_error_list.extend(validation_error)

    error_log = parse_error_list + validate_error_list

    return MorphStatus(file_list, counter, len(parse_error_list), len(validate_error_list), error_log)


@timeit
def morph(workspace_client: WorkspaceClient, config: MorphConfig):
    """
    [Experimental] Transpiles the SQL queries from one dialect to another.

    :param config: The configuration for the morph operation.
    :param workspace_client: The WorkspaceClient object.
    """
    if not config.input_sql:
        logger.error("Input SQL path is not provided.")
        raise ValueError("Input SQL path is not provided.")

    input_sql = Path(config.input_sql)
    status = []
    result = MorphStatus([], 0, 0, 0, [])

    read_dialect = config.get_read_dialect()
    transpiler = SqlglotEngine(read_dialect)
    validator = None
    if not config.skip_validation:
        sql_backend = db_sql.get_sql_backend(workspace_client)
        logger.info(f"SQL Backend used for query validation: {type(sql_backend).__name__}")
        validator = Validator(sql_backend)

    if input_sql.is_file():
        if is_sql_file(input_sql):
            msg = f"Processing for sqls under this file: {input_sql}"
            logger.info(msg)
            if config.output_folder in {None, "None"}:
                output_folder = input_sql.parent / "transpiled"
            else:
                output_folder = Path(str(config.output_folder).rstrip("/"))

            make_dir(output_folder)
            output_file = output_folder / input_sql.name
            no_of_sqls, parse_error, validation_error = _process_file(
                config, validator, transpiler, input_sql, output_file
            )
            error_log = parse_error + validation_error
            result = MorphStatus([str(input_sql)], no_of_sqls, len(parse_error), len(validation_error), error_log)
        else:
            msg = f"{input_sql} is not a SQL file."
            logger.warning(msg)
    elif input_sql.is_dir():
        result = _process_recursive_dirs(config, input_sql, validator, transpiler)
    else:
        msg = f"{input_sql} does not exist."
        logger.error(msg)
        raise FileNotFoundError(msg)

    error_list_count = result.parse_error_count + result.validate_error_count
    if not config.skip_validation:
        logger.info(f"No of Sql Failed while Validating: {result.validate_error_count}")

    error_log_file = "None"
    if error_list_count > 0:
        error_log_file = str(Path.cwd().joinpath(f"err_{os.getpid()}.lst"))
        if result.error_log_list:
            with Path(error_log_file).open("a") as e:
                e.writelines(f"{err}\n" for err in result.error_log_list)

    status.append(
        {
            "total_files_processed": len(result.file_list),
            "total_queries_processed": result.no_of_queries,
            "no_of_sql_failed_while_parsing": result.parse_error_count,
            "no_of_sql_failed_while_validating": result.validate_error_count,
            "error_log_file": str(error_log_file),
        }
    )
    return status


def verify_workspace_client(workspace_client: WorkspaceClient) -> WorkspaceClient:
    # pylint: disable=protected-access
    """
    [Private] Verifies and updates the workspace client configuration.

    TODO: In future refactor this function so it can be used for reconcile module without cross access.
    """
    if workspace_client.config._product != "remorph":
        workspace_client.config._product = "remorph"
    if workspace_client.config._product_version != __version__:
        workspace_client.config._product_version = __version__
    return workspace_client


def _parse(
    transpiler: SqlglotEngine,
    write_dialect: Dialect,
    sql: str,
    input_file: str | Path,
    error_list: list[ParserError],
) -> TranspilationResult:
    return transpiler.transpile(write_dialect, sql, str(input_file), error_list)


def _validation(
    validator: Validator,
    config: MorphConfig,
    sql: str,
) -> ValidationResult:
    return validator.validate_format_result(config, sql)


@timeit
def morph_sql(
    workspace_client: WorkspaceClient,
    config: MorphConfig,
    sql: str,
) -> tuple[TranspilationResult, ValidationResult | None]:
    """[Experimental] Transpile a single SQL query from one dialect to another."""
    ws_client: WorkspaceClient = verify_workspace_client(workspace_client)

    read_dialect: Dialect = config.get_read_dialect()
    write_dialect: Dialect = config.get_write_dialect()
    transpiler: SqlglotEngine = SqlglotEngine(read_dialect)

    transpiler_result = _parse(transpiler, write_dialect, sql, "inline_sql", [])

    if not config.skip_validation:
        sql_backend = db_sql.get_sql_backend(ws_client)
        logger.info(f"SQL Backend used for query validation: {type(sql_backend).__name__}")
        validator = Validator(sql_backend)
        return transpiler_result, _validation(validator, config, transpiler_result.transpiled_sql[0])

    return transpiler_result, None


@timeit
def morph_column_exp(
    workspace_client: WorkspaceClient,
    config: MorphConfig,
    expressions: list[str],
) -> list[tuple[TranspilationResult, ValidationResult | None]]:
    """[Experimental] Transpile a list of SQL expressions from one dialect to another."""
    config.skip_validation = True
    result = []
    for sql in expressions:
        result.append(morph_sql(workspace_client, config, sql))
    return result



from pyspark.sql import SparkSession

from databricks.labs.remorph.reconcile.recon_config import JdbcReaderOptions


class JDBCReaderMixin:
    _spark: SparkSession

    # TODO update the url
    def _get_jdbc_reader(self, query, jdbc_url, driver):
        driver_class = {
            "oracle": "oracle.jdbc.driver.OracleDriver",
            "snowflake": "net.snowflake.client.jdbc.SnowflakeDriver",
        }
        return (
            self._spark.read.format("jdbc")
            .option("url", jdbc_url)
            .option("driver", driver_class.get(driver, driver))
            .option("dbtable", f"({query}) tmp")
        )

    @staticmethod
    def _get_jdbc_reader_options(options: JdbcReaderOptions):
        return {
            "numPartitions": options.number_partitions,
            "partitionColumn": options.partition_column,
            "lowerBound": options.lower_bound,
            "upperBound": options.upper_bound,
            "fetchsize": options.fetch_size,
        }



import json
import os


from databricks.labs.blueprint.cli import App
from databricks.labs.blueprint.entrypoint import get_logger
from databricks.labs.remorph.config import SQLGLOT_DIALECTS, MorphConfig
from databricks.labs.remorph.contexts.application import ApplicationContext
from databricks.labs.remorph.helpers.recon_config_utils import ReconConfigPrompts
from databricks.labs.remorph.reconcile.runner import ReconcileRunner
from databricks.labs.remorph.lineage import lineage_generator
from databricks.labs.remorph.transpiler.execute import morph
from databricks.sdk import WorkspaceClient

remorph = App(__file__)
logger = get_logger(__file__)

DIALECTS = {name for name, dialect in SQLGLOT_DIALECTS.items()}


def raise_validation_exception(msg: str) -> Exception:
    raise ValueError(msg)


@remorph.command
def transpile(
    w: WorkspaceClient,
    source: str,
    input_sql: str,
    output_folder: str | None,
    skip_validation: str,
    catalog_name: str,
    schema_name: str,
    mode: str,
):
    """Transpiles source dialect to databricks dialect"""
    ctx = ApplicationContext(w)
    logger.info(f"User: {ctx.current_user}")
    default_config = ctx.transpile_config
    if not default_config:
        raise SystemExit("Installed transpile config not found. Please install Remorph transpile first.")
    _override_workspace_client_config(ctx, default_config.sdk_config)
    mode = mode if mode else "current"  # not checking for default config as it will always be current
    if source.lower() not in SQLGLOT_DIALECTS:
        raise_validation_exception(f"Error: Invalid value for '--source': '{source}' is not one of {DIALECTS}.")
    if not input_sql or not os.path.exists(input_sql):
        raise_validation_exception(f"Error: Invalid value for '--input_sql': Path '{input_sql}' does not exist.")
    if not output_folder:
        output_folder = default_config.output_folder if default_config.output_folder else None
    if skip_validation.lower() not in {"true", "false"}:
        raise_validation_exception(
            f"Error: Invalid value for '--skip_validation': '{skip_validation}' is not one of 'true', 'false'."
        )
    if mode.lower() not in {"current", "experimental"}:
        raise_validation_exception(
            f"Error: Invalid value for '--mode': '{mode}' " f"is not one of 'current', 'experimental'."
        )

    sdk_config = default_config.sdk_config if default_config.sdk_config else None
    catalog_name = catalog_name if catalog_name else default_config.catalog_name
    schema_name = schema_name if schema_name else default_config.schema_name

    config = MorphConfig(
        source=source.lower(),
        input_sql=input_sql,
        output_folder=output_folder,
        skip_validation=skip_validation.lower() == "true",  # convert to bool
        catalog_name=catalog_name,
        schema_name=schema_name,
        mode=mode,
        sdk_config=sdk_config,
    )

    status = morph(ctx.workspace_client, config)

    print(json.dumps(status))


def _override_workspace_client_config(ctx: ApplicationContext, overrides: dict[str, str] | None):
    """
    Override the Workspace client's SDK config with the user provided SDK config.
    Users can provide the cluster_id and warehouse_id during the installation.
    This will update the default config object in-place.
    """
    if not overrides:
        return

    warehouse_id = overrides.get("warehouse_id")
    if warehouse_id:
        ctx.connect_config.warehouse_id = warehouse_id

    cluster_id = overrides.get("cluster_id")
    if cluster_id:
        ctx.connect_config.cluster_id = cluster_id


@remorph.command
def reconcile(w: WorkspaceClient):
    """[EXPERIMENTAL] Reconciles source to Databricks datasets"""
    ctx = ApplicationContext(w)
    logger.info(f"User: {ctx.current_user}")
    recon_runner = ReconcileRunner(
        ctx.workspace_client,
        ctx.installation,
        ctx.install_state,
        ctx.prompts,
    )
    recon_runner.run()


@remorph.command
def generate_lineage(w: WorkspaceClient, source: str, input_sql: str, output_folder: str):
    """[Experimental] Generates a lineage of source SQL files or folder"""
    ctx = ApplicationContext(w)
    logger.info(f"User: {ctx.current_user}")
    if source.lower() not in SQLGLOT_DIALECTS:
        raise_validation_exception(f"Error: Invalid value for '--source': '{source}' is not one of {DIALECTS}.")
    if not input_sql or not os.path.exists(input_sql):
        raise_validation_exception(f"Error: Invalid value for '--input_sql': Path '{input_sql}' does not exist.")
    if not os.path.exists(output_folder) or output_folder in {None, ""}:
        raise_validation_exception(
            f"Error: Invalid value for '--output-folder': Path '{output_folder}' does not exist."
        )

    lineage_generator(source, input_sql, output_folder)


@remorph.command
def configure_secrets(w: WorkspaceClient):
    """Setup reconciliation connection profile details as Secrets on Databricks Workspace"""
    recon_conf = ReconConfigPrompts(w)

    # Prompt for source
    source = recon_conf.prompt_source()

    logger.info(f"Setting up Scope, Secrets for `{source}` reconciliation")
    recon_conf.prompt_and_save_connection_details()


if __name__ == "__main__":
    remorph()



from sqlglot import expressions as exp, parse, transpile
from sqlglot.dialects.dialect import Dialect
from sqlglot.errors import ErrorLevel, ParseError, TokenError, UnsupportedError
from sqlglot.expressions import Expression

from databricks.labs.remorph.config import TranspilationResult
from databricks.labs.remorph.helpers.file_utils import refactor_hexadecimal_chars
from databricks.labs.remorph.helpers.morph_status import ParserError


class SqlglotEngine:
    def __init__(self, read_dialect: Dialect):
        self.read_dialect = read_dialect

    def transpile(
        self, write_dialect: Dialect, sql: str, file_name: str, error_list: list[ParserError]
    ) -> TranspilationResult:
        try:
            transpiled_sql = transpile(sql, read=self.read_dialect, write=write_dialect, pretty=True, error_level=None)
        except (ParseError, TokenError, UnsupportedError) as e:
            transpiled_sql = [""]
            error_list.append(ParserError(file_name, refactor_hexadecimal_chars(str(e))))

        return TranspilationResult(transpiled_sql, error_list)

    def parse(self, sql: str, file_name: str) -> tuple[list[Expression | None] | None, ParserError | None]:
        expression = None
        error = None
        try:
            expression = parse(sql, read=self.read_dialect, error_level=ErrorLevel.IMMEDIATE)
        except (ParseError, TokenError, UnsupportedError) as e:
            error = ParserError(file_name, str(e))

        return expression, error

    def parse_sql_content(self, sql, file_name):
        parsed_expression, _ = self.parse(sql, file_name)
        if parsed_expression is not None:
            for expr in parsed_expression:
                child = str(file_name)
                if expr is not None:
                    for create in expr.find_all(exp.Create, exp.Insert, exp.Merge, bfs=False):
                        child = self._find_root_tables(create)

                    for select in expr.find_all(exp.Select, exp.Join, exp.With, bfs=False):
                        yield self._find_root_tables(select), child

    @staticmethod
    def _find_root_tables(expression) -> str | None:
        for table in expression.find_all(exp.Table, bfs=False):
            return table.name
        return None
